/*
	One-time sensor network project power meter
	(C) 2007 Thomas Bleeker
	www.madwizard.org
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace PowerMeter
{
    public partial class Form1 : Form
    {
        // Maximum number of samples to receive:
        const int BufferSize = 1000000;

        // Delegate is needed to update the screen because
        // of differnt UI and processing threads
        private delegate void UpdateSceenDelegate();

        // The buffer that will hold all samples
        private byte[] _dataBuffer;

        // Number of bytes currently used in the buffer
        private int _bytesUsed;

        public Form1(string comPort)
        {
            serial.PortName = comPort;

            // Initialize buffer
            _dataBuffer = new byte[BufferSize];
            InitializeComponent();

            // Create a bitmap for the graph
            graph.Image = new Bitmap(graph.ClientSize.Width, graph.ClientSize.Height, PixelFormat.Format24bppRgb);
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // Open the serial port
            serial.Open();
            UpdateScreen();
        }

        // Horizontal resolution of graph
        float HorRes = 0.05f;           // Pixels / sample
        const int BytesPerSec = 5760;   // Number of samples received each second

        // Interval of the horizontal divisions (seconds)
        const float DivInterval = 0.1f;

        // Current selection (in bytes)
        private int _selectStart = - 1;
        private int _selectEnd = -1;

        // Updates the graph
        private void UpdateScreen()
        {
            // Vertical resolution (pixels / step)
            float verRes = 1.0f;

            // Call this function in UI thread if not currently in the UI thread
            if (this.InvokeRequired)
            {
                this.Invoke(new UpdateSceenDelegate(this.UpdateScreen));
                return;
            }
            
            int size;
            lock (this)
            {
                size = _bytesUsed;
            }
            
            // Update maximum scroll position
            hScroll.Maximum = Math.Max(0, size - 1);

            // Get starting position of current view
            int startPos = hScroll.Value;

            // Draw the current view of the graph
            using (Graphics g = Graphics.FromImage(graph.Image))
            {
                g.PageScale = 1.0f;
                g.PageUnit = GraphicsUnit.Pixel;

                // Clear background
                g.FillRectangle(Brushes.White, graph.ClientRectangle);

                // Determine dimensions of view
                int horPixels = graph.ClientRectangle.Width;
                int verPixels = graph.ClientRectangle.Height - 1;
                
                // Calculate the number of bytes currently visibile
                float lastX = -1, lastY = -1;
                int bytesVisible = (int)(horPixels / HorRes);
                if (bytesVisible + startPos > size)
                    bytesVisible = size - startPos;

                // Draw a blue line to indicate the maximum value
                g.DrawLine(Pens.Blue, 0, verPixels - verRes * 255, horPixels, verPixels - verRes * 255);

                // Draw the current selection
                if (_selectEnd != -1 && _selectStart != -1)
                {
                    g.FillRectangle(Brushes.YellowGreen,
                                (_selectStart - startPos) * HorRes, 0,
                                ((_selectEnd - _selectStart)) * HorRes, verPixels + 1);
                }

                // Draw vertical bars every 5mA
                for (int i = 0; i < 25; i += 5)
                {
                    // Calculate mA from measurements
                    // See Atmel datasheets for calculations with the Analog-digital conversion
                    float value = 255 * (i * 5.25f / 1000.0f) / (2.56f/20);
                    value *= verRes;
                    g.DrawLine(Pens.Gray, 0, verPixels - value, horPixels, verPixels - value);
                    g.DrawString(i.ToString() + "mA", SystemFonts.DefaultFont,
                            Brushes.Black, horPixels-40, verPixels - value - 15);
                            
                }

                // Number of horizontal divisions
                int horDivs = (int)(bytesVisible / (BytesPerSec * DivInterval)) + 2;
                
                // Starting time of current view
                float startTime = (float)startPos / BytesPerSec;

                // Starting point of first horizontal divison
                float startDiv = (float)Math.IEEERemainder(startTime, DivInterval);
                
                // Draw edges of selection
                if (_selectEnd != -1)
                    g.DrawLine(Pens.Red, (_selectEnd - startPos) * HorRes,
                                    0, (_selectEnd - startPos) * HorRes, verPixels);
                if (_selectStart != -1)
                    g.DrawLine(Pens.Red, (_selectStart - startPos) * HorRes,
                                    0, (_selectStart - startPos) * HorRes, verPixels);
                
                // Draw horizontal divisions
                for (int i = 0; i < horDivs; i++)
                {
                    float x = (startDiv + i * DivInterval) * BytesPerSec * HorRes;
                    g.DrawLine(Pens.Green, x, 0, x, verPixels);
                    float time = startTime + i * DivInterval;
                    g.DrawString(String.Format("{0:0.0}", time), 
                                  SystemFonts.DefaultFont, Brushes.DarkOliveGreen, x + 2,
                                  verPixels - 20);
                }

                // Draw the actual graph
                for (int i = 0; i < bytesVisible; i++)
                {
                    int byteIdx = startPos + i;

                    float x = HorRes * i;
                    float y = verPixels - (float)_dataBuffer[byteIdx] * verRes;

                    if (lastX != -1)
                    {
                        g.DrawLine(Pens.Black, lastX, lastY, x, y);
                    }

                    lastX = x;
                    lastY = y;
                }
                
                graph.Invalidate();
            }

            // Update labels
            errorLabel.Text = "Errors: " + _errorCount;
            this.Text = "Power Meter (" + _bytesUsed.ToString() + " bytes received)";
        }
        
        // When _even is true, the next byte received is supposed to be a
        // sequence number. Otherwise, the next byte should be a measurment.
        private bool _even = false;

        // Last byte received
        private int _lastByte = -1;

        // Number of errors encountered. It is normal that there are 
        // errors at the start because the datastream needs to be synchronized
        private int _errorCount = 0;

        // Update the data buffer
        private void UpdateBuffer()
        {
            // Atomic operation
            lock (this)
            {
                
                int size = serial.BytesToRead;
                // Each measurement takes one byte but is received
                // as two bytes. So there is actually twice as much
                // space.
                int bytesFree = (_dataBuffer.Length - _bytesUsed) * 2;
                if (bytesFree == 0)
                    return;
                else if (bytesFree < size)
                    size = bytesFree;

                // Read all data in a temporary buffer
                byte[] temp = new byte[size];
                serial.Read(temp, 0, size);

                
                bool error = false;
                int j = _bytesUsed;

                // Process buffer
                for (int i = 0; i < size; i++)
                {
                    // Get the next byte
                    byte b = temp[i];
                    if (_even)
                    {
                        // Sequence number
                        if (_lastByte != -1)
                        {
                            // Sequence number increased by one?
                            if (b != ((_lastByte+1)&0xFF))
                            {
                                error = true;
                            }
                        }
                        _lastByte = b;
                    }
                    else
                    {
                        // Measurement
                        _dataBuffer[j++] = b;
                    }
                    // Alternating sequence number and measurment:
                    _even = !_even;
                }

                if (error)
                {
                    // If there were errors the data is likely 
                    // out of sync. Change _even to its opposite
                    // and hopefully the next time it will be in 
                    // sync.
                    _lastByte = -1;
                    _even = !_even;
                    _errorCount++;
                    
                }
                else
                {
                    // Everything went ok
                    _bytesUsed = j;
                }
            }
            
            UpdateScreen();
        }

        private void serial_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            // Process data
            UpdateBuffer();
        }

        private void hScroll_Scroll(object sender, ScrollEventArgs e)
        {
            // Refresh graph
            UpdateScreen();
        }

        private void hScroll_Scroll(object sender, EventArgs e)
        {
            // Refresh graph
            UpdateScreen();
        }

        private void graph_MouseDown(object sender, MouseEventArgs e)
        {
            // Update selection
            int pos;
            lock(this)
            {
                // Get position clicked on
                pos = (int)((e.X / HorRes) + hScroll.Value);
                if (pos < 0)
                    pos = 0;
                else if (pos >= _bytesUsed)
                    pos = _bytesUsed - 1;
            }

            if (e.Button == MouseButtons.Left)
            {
                // Left button -> start selection
                _selectStart = pos;
            }
            else if (e.Button == MouseButtons.Right)
            {
                // Right button -> end selection
                _selectEnd = pos;

            }
            else if (e.Button == MouseButtons.Middle)
            {
                // Middle button -> select three seconds
                _selectEnd = Math.Min(_bytesUsed - 1, _selectStart + BytesPerSec * 2);
            }

            UpdateScreen();
            
        }

        private void csvButton_Click(object sender, EventArgs e)
        {
            // Export data to CSV

            // Ensure there is a selection
            if (_selectEnd == -1 || _selectStart == -1)
                return;

            // Show a dialog with were to save the file
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                // Create a new file
                using (FileStream str = new FileStream(saveFileDialog.FileName, FileMode.Create))
                {
                    using (StreamWriter wr = new StreamWriter(str))
                    {
                        // Header
                        wr.WriteLine("Time;Current");

                        // Number of mesurements
                        int count = _selectEnd - _selectStart;

                        // Use a scaler to limit the number of measurments
                        // to 1000. It will average the total selection so
                        // that it will be under this number. This ensures that
                        // for example Excel can handle it because it has
                        // problems with large CSV files.
                        int scaler = 1;
                        while ((count / scaler) > 1000)
                            scaler++;

                        // Produce the output
                        for (int i = 0; i < count; i += scaler)
                        {
                            float val = 0;
                            // Average scaler number of items
                            for (int j = 0; j < scaler; j++)
                                val += _dataBuffer[_selectStart + i + j];
                            val /= scaler;

                            // Convert measurment to mA. See Atmel's datasheet
                            // on ADCs for more information
                            val = (val / 255 * (2.56f / 20)) * (1000.0f / 5.25f);

                            // Include a time (starting at 0)
                            float time = ((float)i / BytesPerSec) * 1000;
                            wr.WriteLine(String.Format("{0:0.000};{1:0.000}", time, val));
                        }
                        wr.Close();
                        str.Close();
                    }
                }
            }
        }
    }
}